package convey

import (
	"github.com/smartystreets/goconvey/execution"
	"github.com/smartystreets/goconvey/gotest"
	"github.com/smartystreets/goconvey/printing"
	"github.com/smartystreets/goconvey/reporting"
	"os"
)

// Convey is the method intended for use when declaring the scopes
// of a specification. Each scope has a description and a func()
// which may contain other calls to Convey(), Reset() or Should-style
// assertions. Convey calls can be nested as far as you see fit.
//
// IMPORTANT NOTE: The top-level Convey() within a Test method
// must conform to the following signature:
//
//     Convey(description string, t *testing.T, action func())
//
// All other calls should like like this (no need to pass in *testing.T):
//
//     Convey(description string, action func())
//
// See the examples package for, well, examples.
func Convey(items ...interface{}) {
	name, action, test := parseRegistration(items)
	register(name, action, test)
}

// SkipConvey is analagous to Convey except that the scope is not executed
// (which means that child scopes defined within this scope are not run either).
// The reporter will be notified that this step was skipped.
func SkipConvey(items ...interface{}) {
	name, _, test := parseRegistration(items)
	action := execution.NewSkippedAction(skipReport)
	register(name, action, test)
}

func register(name string, action *execution.Action, test gotest.T) {
	if test != nil {
		runner.Begin(test, name, action)
		runner.Run()
	} else {
		runner.Register(name, action)
	}
}

func skipReport() {
	reporter.Report(reporting.NewSkipReport())
}

// Reset registers a cleanup function to be run after each Convey()
// in the same scope. See the examples package for a simple use case.
func Reset(action func()) {
	runner.RegisterReset(execution.NewAction(action))
}

// So is the means by which assertions are made against the system under test.
// The majority of exported names in this package begin with the word 'Should'
// and describe how the first argument (actual) should compare with any of the
// final (expected) arguments. How many final arguments are accepted depends on
// the particular assertion that is passed in as the assert argument.
// See the examples package for use cases and the assertions package for
// documentation on specific assertion methods.
func So(actual interface{}, assert assertion, expected ...interface{}) {
	if result := assert(actual, expected...); result == assertionSuccess {
		reporter.Report(reporting.NewSuccessReport())
	} else {
		reporter.Report(reporting.NewFailureReport(result))
	}
}

// SkipSo is analagous to So except that the assertion that would have been passed
// to So is not executed and the reporter is notified that the assertion was skipped.
func SkipSo(stuff ...interface{}) {
	skipReport()
}

// assertion is an alias for a function with a signature that the convey.So()
// method can handle. Any future or custom assertions should conform to this
// method signature. The return value should be an empty string if the assertion
// passes and a well-formed failure message if not.
type assertion func(actual interface{}, expected ...interface{}) string

func init() {
	reporter = buildReporter()
	runner = execution.NewRunner()
	runner.UpgradeReporter(reporter)
}
func buildReporter() reporting.Reporter {
	var consoleReporter reporting.Reporter
	console := printing.NewConsole()
	printer := printing.NewPrinter(console)

	if verbose() {
		consoleReporter = reporting.NewStoryReporter(printer)
	} else {
		consoleReporter = reporting.NewDotReporter(printer)
	}

	return reporting.NewReporters(
		reporting.NewGoTestReporter(),
		consoleReporter,
		reporting.NewProblemReporter(printer),
		reporting.NewStatisticsReporter(printer))
}
func verbose() bool {
	for _, arg := range os.Args {
		if arg == verboseEnabled {
			return true
		}
	}
	return false
}

var runner execution.Runner
var reporter reporting.Reporter

const verboseEnabled = "-test.v=true"
const assertionSuccess = ""
